// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

library deequery;

import 'package:darmatch/matchers.dart';
import 'package:darmatch/preconditions.dart';

part 'src/comparables.dart';
part 'src/lazy_query_impl.dart';
part 'src/eager_query_impl.dart';

/// [Query] is a view on top of an [Iterable] that allows declarative
/// (functional) style of working with collections.
abstract class Query<E> extends Iterable<E> {
  /// Returns a concatenation of this collection and the [other] one.
  Query<E> append(Iterable<E> other);

  /// Returns a collection where each object is present only once,
  /// according to the `==` operator. All elements of the collection should
  /// implement it.
  Query<E> unique();

  /// Returns a collection which contains only the elements that matches given
  /// [predicate]. It must be either [Predicate] or [Matcher].
  Query<E> filter(predicate);

  /// Returns a collection which contains only the elements that DON'T match
  /// given [predicate]. It must be either [Predicate] or [Matcher].
  Query<E> reject(predicate);

  /// Sorts the collection according to given [compare] function. It can be
  /// omitted, in which case all elements of the collection must be [Comparable]
  /// and natural sort is used.
  Query<E> sort([int compare(E a, E b)]);

  /// Sorts the collection by the key extracted from each element by the
  /// [extractCompareKey] function. The key must be [Comparable].
  /// The function can also return a [List<Comparable>], in which case
  /// the first element is used as a key and if two keys are equal, the next
  /// element of the list is used, etc.
  Query<E> sortBy(extractCompareKey(E e));

  /// Returns the collection in reverse order.
  Query<E> reverse();

  /// Returns a collection which for each element of this collection contains
  /// a result of calling [transform].
  Query/*<F>*/ map(/*F*/ transform(E e));

  /// Returns a collection which doesn't contain [count] first elements of this
  /// collection.
  Query<E> skip(int count);

  /// Returns a collection which only contains [count] first elements of this
  /// collection.
  Query<E> take(int count);

  /// Iterates from the beginning of this collection and while the
  /// [predicate] holds, skips the iterated elements. First
  /// element that doesn't satisfy the [predicate] stops the iteration
  /// and becomes the first element of the resulting collection.
  /// The [predicate] must be either [Predicate] or [Matcher].
  Query<E> skipWhile(predicate);

  /// Iterates from the beginning of this collection and while the
  /// [predicate] holds, takes the iterated elements (i.e., adds it
  /// to the resulting collection). First element that doesn't satisfy
  /// the [predicate] stops the iteration and isn't part of the resulting
  /// collection. The [predicate] must be either [Predicate] or [Matcher].
  Query<E> takeWhile(predicate);

  /// Finds out if this collection is empty.
  bool isEmpty();

  /// Finds out if this collection contains an element that matches given
  /// [predicate]. It must be either [Predicate] or [Matcher].
  bool some(predicate);

  /// Finds out if all elements of this collection matches given [predicate].
  /// It must be either [Predicate] or [Matcher].
  bool every(predicate);

  /// Finds out if the collection contains an element equal to [e]
  /// according to the `==` operator. All elements of the collection
  /// should implement it.
  bool contains(E e);

  /// Finds the first element of the collection that satisfies given
  /// [predicate] (which can either be a [Predicate] or a [Matcher]).
  E findFirst(predicate);

  /// Combines all elements of the collection by applying the [aggregate]
  /// function. Each element of the collection, beginning with the first
  /// and ending with the last, is passed to the [aggregate] function
  /// together with the accumulator value and the result is used as a new
  /// accumulator. Initial value of the accumulator is [seed].
  /// After all the elements were passed to the [aggregate] function,
  /// the accumulator value is returned. It means that if the collection
  /// is empty, the [seed] value is returned.
  /*F*/ fold(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e));

  /// [foldRight] is like [fold], except that the elements of the collection
  /// are processed in reverse order (from last to first).
  /*F*/ foldRight(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e));

  /// [reduce] is like [fold], except that it doesn't take an extra seed.
  /// First element of the collection is used as a seed instead and only
  /// the rest of the collection is processed. If the collection is empty,
  /// `null` is returned.
  E reduce(E aggregate(E accumulator, E e));

  /// [reduceRight] is like [fold], except that the elements of the collection
  /// are processed in reverse order (from last to first) and it doesn't take
  /// an extra seed. Last element of the collection is used as a seed instead
  /// and only the rest of the collection is processed. If the collection is
  /// empty, `null` is returned.
  E reduceRight(E aggregate(E accumulator, E e));

  /// Finds the maximum element according to the `>` operator. All elements
  /// of the collection must implement it. If the collection is empty,
  /// `null` is returned.
  E max();

  /// Finds the minimum element according to the `<` operator. All elements
  /// of the collection must implement it. If the collection is empty,
  /// `null` is returned.
  E min();

  /// Sums the collection. All elements must implement the `+` operator.
  /// If the collection is empty, `null` is returned.
  E sum();

  /// Joins all objects to one string, separated by the [separator].
  /// By default, `null` values are skipped; you can override that
  /// with the [skipNulls] argument.
  String asString(String separator, {bool skipNulls: true});

  /// Returns a single element of the collection based on the [index].
  /// The [index] can either be `>= 0`, in which case it is the same
  /// as indexing regular collection, or it can be `< 0`, in which case
  /// it means indexing from the end (where `-1` is the last element,
  /// `-2` is the second last etc.). Throws [IndexOutOfRangeException]
  /// if the [index] is out of bounds (i.e., is `>= length` or `< -length`).
  E operator [](int index);
}

/// Queries the [iterable] lazily. That is, only the neccessary elements
/// of the [iterable] are read. This is useful when the [iterable]
/// is computed on the fly and the computation is expensive.
///
/// The [iterable] is read only when the [Query] is iterated or some
/// aggregation function is called, so all changes made to the [iterable]
/// after the [Query] is constructed are reflected (i.e., the [Query]
/// is backed by the [iterable]). You can iterate (or aggregate) the [Query]
/// many times, but each iteration means whole new computation.
///
/// There are some operations that requires access to all elements, in which
/// case the collection is fully read and the query is simply converted
/// to eager one. Those are:
///
/// - `sort`
/// - `sortBy`
/// - `reverse`
/// - `isEmpty`
/// - `every`
/// - `fold`
/// - `foldRight`
/// - `reduce`
/// - `reduceRight`
/// - `max`
/// - `min`
/// - `sum`
/// - `asString`
/// - the `[]` indexing operator, when the index is negative
Query/*<E>*/ query(Iterable/*<E>*/ iterable) => new _LazyQueryImpl/*<E>*/(iterable);

/// Queries the [iterable] eagerly. That is, the [iterable] is fully read
/// at construction time, so it must fit into memory completely. Changes to
/// the original [iterable] after creating the [Query] aren't reflected
/// (i.e., the [Query] isn't backed by the [iterable]).
Query/*<E>*/ eagerQuery(Iterable/*<E>*/ iterable) => new _EagerQueryImpl/*<E>*/(iterable);
