// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

part of deequery;

class _LazyQueryImpl<E> implements Query<E> {
  Iterable<E> _iterable;

  _LazyQueryImpl(this._iterable);

  Iterator<E> iterator() => _iterable.iterator();

  Query<E> get immediate => new _EagerQueryImpl<E>(_iterable);

  Query<E> append(Iterable<E> other) {
    _iterable = new _CompoundIterable(_iterable, other);
    return this;
  }

  Query<E> unique() {
    _iterable = new _UniqueIterable(_iterable);
    return this;
  }

  Query<E> filter(predicate) {
    _iterable = new _FilteredIterable(_iterable, positivePredicate(predicate));
    return this;
  }

  Query<E> reject(predicate) {
    _iterable = new _FilteredIterable(_iterable, negativePredicate(predicate));
    return this;
  }

  Query<E> sort([int compare(E a, E b) = null]) => immediate.sort(compare);

  Query<E> sortBy(extractCompareKey(E e)) => immediate.sortBy(extractCompareKey);

  Query<E> reverse() => immediate.reverse();

  Query/*<F>*/ map(/*F*/ transform(E e)) {
    _iterable = new _MapIterable(_iterable, transform);
    return this;
  }

  Query<E> skip(int count) {
    _iterable = new _SkipIterable(_iterable, count);
    return this;
  }

  Query<E> take(int count) {
    _iterable = new _TakeIterable(_iterable, count);
    return this;
  }

  Query<E> skipWhile(predicate) {
    _iterable = new _SkipWhileIterable(_iterable, positivePredicate(predicate));
    return this;
  }

  Query<E> takeWhile(predicate) {
    _iterable = new _TakeWhileIterable(_iterable, positivePredicate(predicate));
    return this;
  }

  bool isEmpty() => immediate.isEmpty();

  bool some(predicate) {
    final p = positivePredicate(predicate);
    Iterator<E> it = _iterable.iterator();
    while (it.hasNext) {
      E e = it.next();
      if (p(e)) {
        return true;
      }
    }
    return false;
  }

  bool every(predicate) => immediate.every(predicate);

  bool contains(E e) => some( (item) => e == item );

  E findFirst(predicate) {
    final p = positivePredicate(predicate);
    Iterator<E> it = _iterable.iterator();
    while (it.hasNext) {
      E e = it.next();
      if (p(e)) {
        return e;
      }
    }

    return null; // TODO what else?
  }

  /*F*/ fold(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e)) => immediate.fold(seed, aggregate);
  /*F*/ foldRight(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e)) => immediate.foldRight(seed, aggregate);
  E reduce(E aggregate(E accumulator, E e)) => immediate.reduce(aggregate);
  E reduceRight(E aggregate(E accumulator, E e)) => immediate.reduceRight(aggregate);

  E max() => immediate.max();

  E min() => immediate.min();

  E sum() => immediate.sum();

  String asString(String separator, {bool skipNulls: true}) => immediate.asString(separator, skipNulls: skipNulls);

  E operator [](int index) {
    if (index >= 0) {
      var it = iterator();
      while (index > 0 && it.hasNext) {
        it.next();
        index--;
      }
      if (index > 0 || !it.hasNext) {
        throw new RangeError(index);
      }
      return it.next();
    } else {
      return immediate[index];
    }
  }
}

// append

class _CompoundIterator<E> implements Iterator<E> {
  Iterator<E> first, second;
  bool inFirst = true;

  _CompoundIterator(this.first, this.second);

  bool get hasNext {
    if (inFirst) {
      bool result = first.hasNext;
      if (!result) {
        inFirst = false;
        result = second.hasNext;
      }
      return result;
    } else {
      return second.hasNext;
    }
  }

  E next() => inFirst ? first.next() : second.next();
}

class _CompoundIterable<E> implements Iterable<E> {
  Iterable<E> first, second;

  _CompoundIterable(this.first, this.second);

  Iterator<E> iterator() => new _CompoundIterator(first.iterator(), second.iterator());
}

// unique

// I hate it, but it's O(N**2) now...
class _UniqueIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  List<E> alreadySeen = <E>[];

  E nextValue;

  _UniqueIterator(this.iterator);

  bool wasAlreadySeen(E element) => alreadySeen.some( (e) => e == element );

  bool get hasNext {
    if (nextValue != null) {
      return true;
    }

    while (iterator.hasNext) {
      E value = iterator.next();
      if (!wasAlreadySeen(value)) {
        nextValue = value;
        alreadySeen.add(value);
        return true;
      }
    }

    return false;
  }

  E next() {
    E result = nextValue;
    nextValue = null;
    return result;
  }
}

class _UniqueIterable<E> implements Iterable<E> {
  Iterable<E> iterable;

  _UniqueIterable(this.iterable);

  Iterator<E> iterator() => new _UniqueIterator(iterable.iterator());
}

// filter, reject

class _FilteredIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;

  E nextValue;
  bool nextValueFound = false;

  _FilteredIterator(this.iterator, this.predicate);

  bool get hasNext {
    if (nextValueFound) {
      return true;
    }

    while (iterator.hasNext) {
      var next = iterator.next();
      if (predicate(next)) {
        nextValue = next;
        nextValueFound = true;
        return true;
      }
    }

    return false;
  }

  E next() {
    if (!nextValueFound) {
      hasNext;
    }
    if (!nextValueFound) {
      throw new StateError("No more elements");
    }

    nextValueFound = false;
    return nextValue;
  }
}

class _FilteredIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _FilteredIterable(this.iterable, this.predicate);

  Iterator<E> iterator() => new _FilteredIterator(iterable.iterator(), predicate);
}

// map

class _MapIterator<E, F> implements Iterator<F> {
  Iterator<E> iterator;
  var transform;

  _MapIterator(this.iterator, F this.transform(E e));

  bool get hasNext => iterator.hasNext;

  F next() => transform(iterator.next());
}

class _MapIterable<E, F> implements Iterable<F> {
  Iterable<E> iterable;
  var transform;

  _MapIterable(this.iterable, F this.transform(E e));

  Iterator<F> iterator() => new _MapIterator<E, F>(iterable.iterator(), transform);
}

// skip

class _SkipIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  int toSkip;

  _SkipIterator(this.iterator, this.toSkip);

  bool get hasNext {
    while (toSkip > 0) {
      if (iterator.hasNext) {
        iterator.next();
      }
      toSkip--;
    }

    return iterator.hasNext;
  }

  E next() {
    return iterator.next();
  }
}

class _SkipIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  int count;

  _SkipIterable(this.iterable, this.count);

  Iterator<E> iterator() => new _SkipIterator(iterable.iterator(), count);
}

// take

class _TakeIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  int toTake;

  _TakeIterator(this.iterator, this.toTake);

  bool get hasNext => toTake > 0 && iterator.hasNext;

  E next() {
    toTake--;
    return iterator.next();
  }
}

class _TakeIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  int count;

  _TakeIterable(this.iterable, this.count);

  Iterator<E> iterator() => new _TakeIterator(iterable.iterator(), count);
}

// skip while

class _SkipWhileIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;
  bool skipped = false;
  E firstValue;

  _SkipWhileIterator(this.iterator, this.predicate);

  bool get hasNext {
    if (!skipped) {
      while (iterator.hasNext) {
        var e = iterator.next();
        if (!predicate(e)) {
          firstValue = e;
          break;
        }
      }
      skipped = true;
    }

    if (firstValue != null) {
      return true;
    }
    return iterator.hasNext;
  }

  E next() {
    if (firstValue != null) {
      var result = firstValue;
      firstValue = null;
      return result;
    }

    return iterator.next();
  }
}

class _SkipWhileIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _SkipWhileIterable(this.iterable, this.predicate);

  Iterator<E> iterator() => new _SkipWhileIterator(iterable.iterator(), predicate);
}

// take while

class _TakeWhileIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;
  bool taken;
  E nextValue;

  _TakeWhileIterator(this.iterator, this.predicate) : taken = false;

  bool get hasNext {
    if (taken) {
      return false;
    }

    if (iterator.hasNext) {
      var e = iterator.next();
      var holds = predicate(e);
      nextValue = holds ? e : null;
      if (!holds) {
        taken = true;
      }
      return holds;
    }

    return false;
  }

  E next() {
    if (nextValue == null) {
      hasNext;
    }

    var result = nextValue;
    nextValue = null;
    return result;
  }
}

class _TakeWhileIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _TakeWhileIterable(this.iterable, this.predicate);

  Iterator<E> iterator() => new _TakeWhileIterator(iterable.iterator(), predicate);
}
