// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

import 'package:deequery/deequery.dart';
import 'package:darmatch/matchers.dart';

import 'package:detester/detester.dart';
part 'package:detester/test.dart';

allTests() => [
  testAppend,
  testUnique,
  testFilter,
  testReject,
  testSort,
  testSortBy,
  testReverse,
  testMap,
  testSkip,
  testTake,
  testSkipWhile,
  testTakeWhile,
  testIsEmpty,
  testSome,
  testEvery,
  testContains,
  testFindFirst,
  testFold,
  testFoldRight,
  testReduce,
  testReduceRight,
  testMax,
  testMin,
  testSum,
  testAsString,
  testIndexing
];

suite() => multi([
  configure("Lazy query", allTests(),
    beforeSuite: () { q = query; }
  ),

  configure("Eager query", allTests(),
    beforeSuite: () { q = eagerQuery; }
  )
]);

var q;

testAppend() {
  var c1 = [1, 2, 3];
  var c2 = [2, 1, 0];
  expectThat(q(c1).append(c2), orderedEquals([1, 2, 3, 2, 1, 0]));
}

testUnique() {
  var c = [1, 2, 4, 1, 3, 4, 0, 1];
  expectThat(q(c).unique(), orderedEquals([1, 2, 4, 3, 0]));
}

testFilter() {
  var c = [1, 2, 3, 4, 5];
  expectThat(q(c).filter(even()), orderedEquals([2, 4]));
}

testReject() {
  var c = [1, 2, 3, 4, 5];
  expectThat(q(c).reject(even()), orderedEquals([1, 3, 5]));
}

testSort() {
  var c = [1, 5, 2, 6, 3, 4, 2, 0, 6];
  expectThat(q(c).sort(), orderedEquals([0, 1, 2, 2, 3, 4, 5, 6, 6]));
  expectThat(q(c).sort( (a, b) => b.compareTo(a) ), orderedEquals([6, 6, 5, 4, 3, 2, 2, 1, 0]));
}

class Employee {
  String name, department;
  num grade;

  Employee(this.name, this.department, this.grade);

  String toString() => "$name (department: $department, grade: $grade)";
}

testSortBy() {
  var employees = [
    new Employee("John", "Marketing", 2.4),
    new Employee("Mary", "Marketing", 2.4),
    new Employee("Gerald", "Sales", 3.6),
    new Employee("Antony", "Administration", 4.2),
    new Employee("Peter", "Administration", 4.8)
  ];

  var sorted = q(employees).sortBy( (e) => [e.department, desc(e.grade), e.name] );

  expectThat(q(sorted).map( (e) => "$e" ), orderedEquals([
    "Peter (department: Administration, grade: 4.8)",
    "Antony (department: Administration, grade: 4.2)",
    "John (department: Marketing, grade: 2.4)",
    "Mary (department: Marketing, grade: 2.4)",
    "Gerald (department: Sales, grade: 3.6)"
  ]));
}

testReverse() {
  var c = [1, 5, 3, 0, 2, 4];
  expectThat(q(c).reverse(), orderedEquals([4, 2, 0, 3, 5, 1]));
}

testMap() {
  var c = [1, 2, 3, 4, 5];
  expectThat(q(c).map( (e) => "${e * 2}" ), orderedEquals(["2", "4", "6", "8", "10"]));
}

testSkip() {
  var c = [1, 2, 3];
  expectThat(q(c).skip(0), orderedEquals([1, 2, 3]));
  expectThat(q(c).skip(1), orderedEquals([2, 3]));
  expectThat(q(c).skip(2), orderedEquals([3]));
  expectThat(q(c).skip(3), orderedEquals([]));
  expectThat(q(c).skip(4), orderedEquals([]));
}

testTake() {
  var c = [1, 2, 3];
  expectThat(q(c).take(4), orderedEquals([1, 2, 3]));
  expectThat(q(c).take(3), orderedEquals([1, 2, 3]));
  expectThat(q(c).take(2), orderedEquals([1, 2]));
  expectThat(q(c).take(1), orderedEquals([1]));
  expectThat(q(c).take(0), orderedEquals([]));
}

testSkipWhile() {
  var c = [1, 2, 3, 2, 1];
  expectThat(q(c).skipWhile(odd()), orderedEquals([2, 3, 2 ,1]));
  expectThat(q(c).skipWhile( (i) => i <= 2 ), orderedEquals([3, 2, 1]));
}

testTakeWhile() {
  var c = [1, 2, 3, 2, 1];
  expectThat(q(c).takeWhile(odd()), orderedEquals([1]));
  expectThat(q(c).takeWhile( (i) => i <= 2 ), orderedEquals([1, 2]));
}

testIsEmpty() {
  expectThat(q([]).isEmpty(), isTrue());
  expectThat(q([1]).isEmpty(), isFalse());
}

testSome() {
  var c = [1, 2, 3, 4];
  expectThat(q(c).some(even()), isTrue());
  expectThat(q(c).some(negative()), isFalse());
}

testEvery() {
  var c = [1, 2, 3, 4];
  expectThat(q(c).every(positive()), isTrue());
  expectThat(q(c).every(even()), isFalse());
}

testContains() {
  var c = [1, 2, 3, 4];
  expectThat(q(c).contains(2), isTrue());
  expectThat(q(c).contains(5), isFalse());
}

testFindFirst() {
  var c = [1, 2, 3, 4];
  expectThat(q(c).findFirst(greaterThan(3)), equals(4));
  expectThat(q(c).findFirst(greaterThan(4)), isNull());
}

testFold() {
  var c = ["a", "b", "c"];
  expectThat(q(c).fold("", (acc, e) => "$acc, $e" ), equals(", a, b, c"));

  expectThat(q([]).fold(0, (acc, e) => acc + e ), equals(0));
  expectThat(q([1]).fold(0, (acc, e) => acc + e ), equals(1));
  expectThat(q([1, 2]).fold(0, (acc, e) => acc + e ), equals(3));
}

testFoldRight() {
  var c = ["a", "b", "c"];
  expectThat(q(c).foldRight("", (acc, e) => "$acc, $e" ), equals(", c, b, a"));

  expectThat(q([]).foldRight(0, (acc, e) => acc + e ), equals(0));
  expectThat(q([1]).foldRight(0, (acc, e) => acc + e ), equals(1));
  expectThat(q([1, 2]).foldRight(0, (acc, e) => acc + e ), equals(3));
}

testReduce() {
  var c = ["a", "b", "c"];
  expectThat(q(c).reduce( (acc, e) => "$acc, $e" ), equals("a, b, c"));

  expectThat(q([]).reduce( (acc, e) => acc + e ), isNull());
  expectThat(q([1]).reduce( (acc, e) => acc + e ), equals(1));
  expectThat(q([1, 2]).reduce( (acc, e) => acc + e ), equals(3));
}

testReduceRight() {
  var c = ["a", "b", "c"];
  expectThat(q(c).reduceRight( (acc, e) => "$acc, $e" ), equals("c, b, a"));

  expectThat(q([]).reduceRight( (acc, e) => acc + e ), isNull());
  expectThat(q([1]).reduceRight( (acc, e) => acc + e ), equals(1));
  expectThat(q([1, 2]).reduceRight( (acc, e) => acc + e ), equals(3));
}

testMax() {
  var c = [1, 4, 2, 5, 3];
  expectThat(q(c).max(), equals(5));

  expectThat(q([]).max(), isNull());
}

testMin() {
  var c = [1, 4, 2, 5, 3];
  expectThat(q(c).min(), equals(1));

  expectThat(q([]).min(), isNull());
}

testSum() {
  var c = [1, 4, 2, 5, 3];
  expectThat(q(c).sum(), equals(15));

  expectThat(q([]).sum(), isNull());
}

testAsString() {
  var c = [1, 2, 3, 4, 5];
  expectThat(q(c).asString(", "), equals("1, 2, 3, 4, 5"));
  expectThat(q(c).map( (i) => i * 2 ).asString(", "), equals("2, 4, 6, 8, 10"));
  expectThat(q(c).map( (i) => "${i * 2}" ).asString(", "), equals("2, 4, 6, 8, 10"));

  c = [1, 2, null, 3, 4];
  expectThat(q(c).asString(", "), equals("1, 2, 3, 4"));
  expectThat(q(c).asString(", ", skipNulls: false), equals("1, 2, null, 3, 4"));

  expectThat(q(c).map( (i) => i === null ? "X" : "${i * 2}" ).asString(", "), equals("2, 4, X, 6, 8"));
  expectThat(q(c).map( (i) => i === null ? null : "${i * 2}" ).asString(", ", skipNulls: false), equals("2, 4, null, 6, 8"));
}

testIndexing() {
  var c = [1, 2, 3];
  expectThat(q(c)[0], equals(1));
  expectThat(q(c)[1], equals(2));
  expectThat(q(c)[2], equals(3));
  expectThat(q(c)[-1], equals(3));
  expectThat(q(c)[-2], equals(2));
  expectThat(q(c)[-3], equals(1));

  expectThat( () => q(c)[3], throws());
  expectThat( () => q(c)[4], throws());
  expectThat( () => q(c)[-4], throws());
  expectThat( () => q(c)[-5], throws());
}
